compoundsuffix.vert.glsl
Shader version: 100
0:? Sequence
0:1  Function Definition: main( ( global void)
0:1    Function Parameters: 
0:3    Sequence
0:3      move second child to first child ( temp highp 4-component vector of float)
0:3        'gl_Position' ( gl_Position highp 4-component vector of float Position)
0:3        Constant:
0:3          1.000000
0:3          1.000000
0:3          1.000000
0:3          1.000000
0:?   Linker Objects

