// Tencent is pleased to support the open source community by making ncnn available.
//
// Copyright (C) 2020 THL A29 Limited, a Tencent company. All rights reserved.
//
// Licensed under the BSD 3-Clause License (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
//
// https://opensource.org/licenses/BSD-3-Clause
//
// Unless required by applicable law or agreed to in writing, software distributed
// under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
// CONDITIONS OF ANY KIND, either express or implied. See the License for the
// specific language governing permissions and limitations under the License.

#include "ncnn_dialect.h"

#include <mlir/IR/Builders.h>

namespace mlir {

namespace ncnn {

NCNNDialect::NCNNDialect(mlir::MLIRContext* context)
    : mlir::Dialect("ncnn", context, TypeID::get<NCNNDialect>())
{
    addOperations<
#define GET_OP_LIST
#include "ncnn_ops.cc.inc"
    >();

    // Support unknown operations because not all NCNN operations are
    // registered.
    allowUnknownOperations();
}

} // namespace ncnn

#define GET_OP_CLASSES
#include "ncnn_ops.cc.inc"

} // namespace mlir
