#define zeros 0
#define zeros1 0
#define zeros2 0, 0
#define zeros3 0, 0, 0
#define zeros4 0, 0, 0, 0
#define zeros5 0, 0, 0, 0, 0
#define zeros6 0, 0, 0, 0, 0, 0
#define zeros7 0, 0, 0, 0, 0, 0, 0
#define zeros8 0, 0, 0, 0, 0, 0, 0, 0
#define zeros9 0, 0, 0, 0, 0, 0, 0, 0, 0
#define zeros10 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
#define zeros11 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
#define zeros12 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
#define zeros13 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
#define zeros14 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
#define zeros15 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
#define zeros16 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
float4 main() : SV_Target {
  float var0 = float(zeros1);
  float2 var13 = float(zeros1);
  float2 var14 = float2(zeros2);
  float3 var26 = float(zeros1);
  float3 var28 = float3(zeros3);
  float4 var39 = float(zeros1);
  float4 var42 = float4(zeros4);
  float4 var43 = float2x2(zeros4);
  float2x2 var52 = float(zeros1);
  float2x2 var55 = float4(zeros4);
  float2x2 var56 = float2x2(zeros4);
  float2x3 var65 = float(zeros1);
  float2x3 var70 = float2x3(zeros6);
  float2x4 var78 = float(zeros1);
  float2x4 var84 = float2x4(zeros8);
  float3x2 var91 = float(zeros1);
  float3x2 var98 = float3x2(zeros6);
  float3x3 var104 = float(zeros1);
  float3x3 var112 = float3x3(zeros9);
  float3x4 var117 = float(zeros1);
  float3x4 var126 = float3x4(zeros12);
  float4x2 var130 = float(zeros1);
  float4x2 var140 = float4x2(zeros8);
  float4x3 var143 = float(zeros1);
  float4x3 var154 = float4x3(zeros12);
  float4x4 var156 = float(zeros1);
  float4x4 var168 = float4x4(zeros16);
  float var1 = float2(zeros2);// warning X3206: implicit truncation of vector type
  float var2 = float3(zeros3);// warning X3206: implicit truncation of vector type
  float var3 = float4(zeros4);// warning X3206: implicit truncation of vector type
  float var4 = float2x2(zeros4);// warning X3206: implicit truncation of vector type
  float var5 = float2x3(zeros6);// warning X3206: implicit truncation of vector type
  float var6 = float2x4(zeros8);// warning X3206: implicit truncation of vector type
  float var7 = float3x2(zeros6);// warning X3206: implicit truncation of vector type
  float var8 = float3x3(zeros9);// warning X3206: implicit truncation of vector type
  float var9 = float3x4(zeros12);// warning X3206: implicit truncation of vector type
  float var10 = float4x2(zeros8);// warning X3206: implicit truncation of vector type
  float var11 = float4x3(zeros12);// warning X3206: implicit truncation of vector type
  float var12 = float4x4(zeros16);// warning X3206: implicit truncation of vector type
  float2 var15 = float3(zeros3);// warning X3206: implicit truncation of vector type
  float2 var16 = float4(zeros4);// warning X3206: implicit truncation of vector type
  float3 var29 = float4(zeros4);// warning X3206: implicit truncation of vector type
  float2x2 var57 = float2x3(zeros6);// warning X3206: implicit truncation of vector type
  float2x2 var58 = float2x4(zeros8);// warning X3206: implicit truncation of vector type
  float2x2 var59 = float3x2(zeros6);// warning X3206: implicit truncation of vector type
  float2x2 var60 = float3x3(zeros9);// warning X3206: implicit truncation of vector type
  float2x2 var61 = float3x4(zeros12);// warning X3206: implicit truncation of vector type
  float2x2 var62 = float4x2(zeros8);// warning X3206: implicit truncation of vector type
  float2x2 var63 = float4x3(zeros12);// warning X3206: implicit truncation of vector type
  float2x2 var64 = float4x4(zeros16);// warning X3206: implicit truncation of vector type
  float2x3 var71 = float2x4(zeros8);// warning X3206: implicit truncation of vector type
  float2x3 var73 = float3x3(zeros9);// warning X3206: implicit truncation of vector type
  float2x3 var74 = float3x4(zeros12);// warning X3206: implicit truncation of vector type
  float2x3 var76 = float4x3(zeros12);// warning X3206: implicit truncation of vector type
  float2x3 var77 = float4x4(zeros16);// warning X3206: implicit truncation of vector type
  float2x4 var87 = float3x4(zeros12);// warning X3206: implicit truncation of vector type
  float2x4 var90 = float4x4(zeros16);// warning X3206: implicit truncation of vector type
  float3x2 var99 = float3x3(zeros9);// warning X3206: implicit truncation of vector type
  float3x2 var100 = float3x4(zeros12);// warning X3206: implicit truncation of vector type
  float3x2 var101 = float4x2(zeros8);// warning X3206: implicit truncation of vector type
  float3x2 var102 = float4x3(zeros12);// warning X3206: implicit truncation of vector type
  float3x2 var103 = float4x4(zeros16);// warning X3206: implicit truncation of vector type
  float3x3 var113 = float3x4(zeros12);// warning X3206: implicit truncation of vector type
  float3x3 var115 = float4x3(zeros12);// warning X3206: implicit truncation of vector type
  float3x3 var116 = float4x4(zeros16);// warning X3206: implicit truncation of vector type
  float3x4 var129 = float4x4(zeros16);// warning X3206: implicit truncation of vector type
  float4x2 var141 = float4x3(zeros12);// warning X3206: implicit truncation of vector type
  float4x2 var142 = float4x4(zeros16);// warning X3206: implicit truncation of vector type
  float4x3 var155 = float4x4(zeros16);// warning X3206: implicit truncation of vector type
  return 0;
}

