#version 460 core
#extension GL_EXT_shader_explicit_arithmetic_types_int64 : enable
#extension GL_EXT_shader_explicit_arithmetic_types_int8 : enable
#extension GL_EXT_shader_explicit_arithmetic_types_int16 : enable
#extension GL_EXT_shader_explicit_arithmetic_types_float16 : enable

int f(int8_t i8,
       uint8_t u8,
       int16_t i16,
       uint16_t u16,
       int64_t i64,
       uint64_t u64,
       float16_t f16)
{
    return 0;
}

void main()
{
}
