#version 310 es

#extension GL_EXT_geometry_shader: enable
#extension GL_EXT_frag_depth: disable
#extension GL_EXT_gpu_shader5: require
#extension GL_EXT_shader_texture_image_samples: warn

#extension unknown_extension: require

int main() {
}

